@@ -6,20 +6,21 @@ from __future__ import unicode_literals
 from prompt_toolkit.enums import DEFAULT_BUFFER, SEARCH_BUFFER
 from prompt_toolkit.filters import IsDone, HasCompletions, RendererHeightIsKnown, Always, HasFocus, Condition
 from prompt_toolkit.key_binding.vi_state import InputMode
-from prompt_toolkit.layout import Window, HSplit, VSplit, FloatContainer, Float, ConditionalContainer
+from prompt_toolkit.layout.containers import Window, HSplit, VSplit, FloatContainer, Float, ConditionalContainer, ScrollOffsets
 from prompt_toolkit.layout.controls import BufferControl, TokenListControl, FillControl
 from prompt_toolkit.layout.dimension import LayoutDimension
 from prompt_toolkit.layout.lexers import SimpleLexer
-from prompt_toolkit.layout.margins import ConditionalMargin, NumberredMargin
+from prompt_toolkit.layout.margins import Margin
 from prompt_toolkit.layout.menus import CompletionsMenu, MultiColumnCompletionsMenu
-from prompt_toolkit.layout.processors import HighlightSearchProcessor, HighlightSelectionProcessor, HighlightMatchingBracketProcessor, ConditionalProcessor
+from prompt_toolkit.layout.processors import HighlightSearchProcessor, HighlightSelectionProcessor, HighlightMatchingBracketProcessor, ConditionalProcessor, AppendAutoSuggestion
 from prompt_toolkit.layout.screen import Char
 from prompt_toolkit.layout.toolbars import CompletionsToolbar, ArgToolbar, SearchToolbar, ValidationToolbar, SystemToolbar, TokenListToolbar
 from prompt_toolkit.layout.utils import token_list_width
 from prompt_toolkit.reactive import Integer
 from prompt_toolkit.selection import SelectionType
+from prompt_toolkit.utils import get_cwidth
 
-from ptpython.filters import HasSignature, ShowSidebar, ShowLineNumbersFilter, ShowSignature, ShowDocstring
+from ptpython.filters import HasSignature, ShowSidebar, ShowSignature, ShowDocstring
 
 from pygments.lexers import PythonLexer
 from pygments.token import Token
@@ -96,13 +97,20 @@ def python_sidebar(python_input):
 
         return tokens
 
+    class Control(TokenListControl):
+        def move_cursor_down(self, cli):
+            python_input.selected_option_index += 1
+
+        def move_cursor_up(self, cli):
+            python_input.selected_option_index -= 1
+
     return ConditionalContainer(
         content=Window(
-            TokenListControl(get_tokens, Char(token=Token.Sidebar),
+            Control(get_tokens, Char(token=Token.Sidebar),
                 has_focus=ShowSidebar(python_input) & ~IsDone()),
             width=LayoutDimension.exact(43),
             height=LayoutDimension(min=3),
-            scroll_offset=1),
+            scroll_offsets=ScrollOffsets(top=1, bottom=1)),
         filter=ShowSidebar(python_input) & ~IsDone())
 
 
@@ -228,15 +236,42 @@ def signature_toolbar(python_input):
             ~IsDone())
 
 
-def python_prompt(python_input):
+class PromptMargin(Margin):
     """
-    Create layout for the prompt.
+    Create margin that displays the prompt.
     It shows something like "In [1]:".
     """
-    def get_tokens(cli):
-        return python_input.get_input_prompt_tokens(cli)
+    def __init__(self, python_input):
+        self.python_input = python_input
 
-    return TokenListControl(get_tokens)
+    def _get_prompt_style(self):
+        return self.python_input.all_prompt_styles[self.python_input.prompt_style]
+
+    def get_width(self, cli):
+        # Take the width from the first line.
+        text = ''.join(t[1] for t in self.python_input.get_input_prompt_tokens(cli))
+        return get_cwidth(text)
+
+    def create_margin(self, cli, window_render_info, width, height):
+        style = self._get_prompt_style()
+
+        # First line.
+        tokens = style.in_tokens(cli)
+
+        # Next lines. (Show line numbering when numbering is enabled.)
+        tokens2 = style.in2_tokens(cli, width)
+        show_numbers = self.python_input.show_line_numbers
+        visible_line_to_input_line = window_render_info.visible_line_to_input_line
+
+        for y in range(1, min(window_render_info.content_height, height)):
+            tokens.append((Token, '\n'))
+            if show_numbers:
+                line_number = visible_line_to_input_line.get(y) or 0
+                tokens.append((Token.LineNumber, ('%i ' % (line_number + 1)).rjust(width)))
+            else:
+                tokens.extend(tokens2)
+
+        return tokens
 
 
 def status_bar(key_bindings_manager, python_input):
@@ -430,9 +465,6 @@ def create_layout(python_input, key_bindings_manager,
             BufferControl(
                 buffer_name=DEFAULT_BUFFER,
                 lexer=lexer,
-                margin=ConditionalMargin(
-                    NumberredMargin(),
-                    filter=ShowLineNumbersFilter(python_input)),
                 input_processors=[
                                   # Show matching parentheses, but only while editing.
                                   ConditionalProcessor(
@@ -442,12 +474,21 @@ def create_layout(python_input, key_bindings_manager,
                                   ConditionalProcessor(
                                       processor=HighlightSearchProcessor(preview_search=Always()),
                                       filter=HasFocus(SEARCH_BUFFER)),
-                                  HighlightSelectionProcessor()] + extra_buffer_processors,
+                                  HighlightSelectionProcessor(),
+                                  ConditionalProcessor(
+                                      processor=AppendAutoSuggestion(),
+                                      filter=~IsDone())
+                                  ] + extra_buffer_processors,
                 menu_position=menu_position,
+                wrap_lines=Condition(lambda cli: python_input.wrap_lines),
 
                 # Make sure that we always see the result of an reverse-i-search:
                 preview_search=Always(),
             ),
+            left_margins=[PromptMargin(python_input)],
+            # Scroll offsets. The 1 at the bottom is important to make sure the
+            # cursor is never below the "Press [Meta+Enter]" message which is a float.
+            scroll_offsets=ScrollOffsets(bottom=1, left=4, right=4),
             # As long as we're editing, prefer a minimal height of 6.
             get_height=(lambda cli: (
                 None if cli.is_done or python_input.show_exit_confirmation
@@ -458,17 +499,9 @@ def create_layout(python_input, key_bindings_manager,
         VSplit([
             HSplit([
                 FloatContainer(
-                    content=HSplit([
-                        VSplit([
-                            Window(
-                                python_prompt(python_input),
-                                dont_extend_width=True,
-                                height=D.exact(1),
-                            ),
-                            create_python_input_window(),
-                        ]),
-                        ] + extra_body + [
-                    ]),
+                    content=HSplit(
+                        [create_python_input_window()] + extra_body
+                    ),
                     floats=[
                         Float(xcursor=True,
                               ycursor=True,
